@@ -1,4 +1,4 @@
-//Copyright (C) 1997-2001 ZSNES Team ( zsknight@zsnes.com / _demo_@zsnes.com )
+//Copyright (C) 1997-2004 ZSNES Team ( zsknight@zsnes.com / _demo_@zsnes.com )
 //
 //This program is free software; you can redistribute it and/or
 //modify it under the terms of the GNU General Public License
@@ -32,13 +32,17 @@ extern void makeextension(void);
 
 extern unsigned char Palette0, SPC700sh, OffBy1Line, DSPDisable,
                      MMXSupport, Force8b, ForcePal, GUIClick, MouseDis,
-                     MusicRelVol, ScreenScale, SoundCompD, SoundQuality,
+                     MusicRelVol, ScreenScale, SoundQuality,
                      StereoSound, V8Mode, antienab, cvidmode, debugdisble,
                      debugger, enterpress, finterleave, frameskip,
                      gammalevel, guioff, per2exec, pl1contrl, pl2contrl,
                      romtype, scanlines, showallext, smallscreenon, soundon,
                      spcon, vsyncon, DisplayS, fname, filefound, SnowOn,
                      Triplebufen;
+                     
+// FIX STATMAT
+extern unsigned char autoloadstate;
+// FIX STATMAT                     
 
 void ccmdline(void);
 
@@ -310,7 +314,15 @@ int pccmdline(void)
         }
         case 'z':
         {
-          StereoSound=1;
+          // FIX STATMAT
+          if(nn == 's')
+          {				
+            if(!hasroom) return 4;
+            autoloadstate=my_atoi(argv[p+1]) + 1;
+            p++;
+          }
+          else StereoSound=0;
+          // FIX STATMAT
           break;
         }
         case 'd':
@@ -327,12 +339,6 @@ int pccmdline(void)
           }
           break;
         }
-        case 'b':
-        {
-          SoundCompD=1;
-          break;
-        }
-
         case 'c':
         {
           if(nn == 'c')
