@@ -21,6 +21,11 @@
 #include "sw_draw.h"
 #include "gl_draw.h"
 
+#include "SDL_thread.h"
+
+#include <sys/time.h>
+#include <time.h>
+
 typedef unsigned char BYTE;
 typedef unsigned short WORD;
 typedef unsigned long DWORD;
@@ -32,11 +37,13 @@ typedef enum { FALSE = 0, TRUE = 1 } BOOL;
 typedef enum vidstate_e { vid_null, vid_none, vid_soft, vid_gl } vidstate_t;
 
 // SOUND RELATED VARIABLES
+SDL_AudioSpec audiospec;
 int SoundEnabled = 1;
 BYTE PrevStereoSound;
 DWORD PrevSoundQuality;
-DWORD BufferLeftOver = 0;	/* should we clear these on sound reset? */
-short Buffer[1800 * 2];
+Uint8 *Buffer = NULL;
+int Buffer_len = 0, Buffer_fill = 0;
+int Buffer_head = 0, Buffer_tail = 0;
 
 extern BYTE StereoSound;
 extern DWORD SoundQuality;
@@ -60,12 +67,32 @@ static int UseOpenGL = 0;
 static const int BitDepth = 16;
 DWORD FirstVid = 1;
 
-extern BYTE GUIWFVID[];
+extern void SwitchFullScreen (void);
 extern unsigned char cvidmode;
+DWORD SMode=0;
+DWORD DSMode=0;
+DWORD prevHQMode=-1;
+
+extern DWORD converta;
+extern DWORD *BitConv32Ptr;
+extern DWORD *RGBtoYUVPtr;
+extern unsigned char hqFilter;
+
+extern BYTE GUIWFVID[];
+extern BYTE GUISMODE[];
+extern BYTE GUIDSMODE[];
+extern BYTE GUIHQ2X[];
+extern BYTE GUIHQ3X[];
+extern BYTE GUIHQ4X[];
 
 /* JOYSTICK AND KEYBOARD INPUT */
 SDL_Joystick *JoystickInput[5];
+unsigned int AxisOffset[5] = {256 + 128 + 64};	// per joystick offsets in
+unsigned int ButtonOffset[5] = {448};		// pressed. We have 128 + 64
+unsigned int HatOffset[5] = {448};		// bytes for all joysticks. We
+unsigned int BallOffset[5] = {448};		// can control all 5 players.
 int shiftptr = 0;
+int offset;
 DWORD numlockptr;
 
 extern unsigned char pressed[];
@@ -91,7 +118,7 @@ extern BYTE GUIOn2;
 static BYTE IsActivated = 1;
 
 /* TIMER VARIABLES/MACROS */
-#define UPDATE_TICKS_GAME (1000.0/59.95)// milliseconds per world update
+#define UPDATE_TICKS_GAME (1000.0/59.94)// milliseconds per world update
 #define UPDATE_TICKS_GAMEPAL (20)	// milliseconds per world update
 #define UPDATE_TICKS_GUI (1000.0/36.0)	// milliseconds per world update
 #define UPDATE_TICKS_UDP (1000.0/60.0)	// milliseconds per world update
@@ -101,9 +128,18 @@ int T36HZEnabled = 0;
 short SystemTimewHour;
 short SystemTimewMinute;
 short SystemTimewSecond;
-Uint32 end, end2;
-double start, start2;
-double update_ticks_pc, update_ticks_pc2;
+float end, end2;
+float start, start2;
+float update_ticks_pc, update_ticks_pc2;
+
+static SDL_sem *sem_frames = NULL;
+
+void sem_sleep_rdy(void);
+void sem_sleep_die(void);
+
+static struct timeval sem_start;
+// void sem_StartTicks(void);
+float sem_GetTicks(void);
 
 extern unsigned char romispal;
 
@@ -118,6 +154,10 @@ void UpdateSound(void *userdata, Uint8 * stream, int len);
 extern int GUI36hzcall(void);
 extern int Game60hzcall(void);
 extern void SoundProcess();
+_int64 copymaskRB = 0x001FF800001FF800LL;
+_int64 copymaskG = 0x0000FC000000FC00LL;
+_int64 copymagic = 0x0008010000080100LL;
+_int64 coef = 0x0066009a0066009aLL;
 #ifdef __OPENGL__
 extern void gl_clearwin(void);
 #endif
@@ -134,7 +174,6 @@ static void adjustMouseYScale(void)
 
 int Main_Proc(void)
 {
-	int j;
 	SDL_Event event;
 
 	while (SDL_PollEvent(&event))
@@ -145,6 +184,11 @@ int Main_Proc(void)
 				IsActivated = event.active.gain;
 				break;
 			case SDL_KEYDOWN:
+				if ((event.key.keysym.sym == SDLK_RETURN) &&
+				    (event.key.keysym.mod & KMOD_ALT)) {
+					SwitchFullScreen();
+					break;
+				}
 				if (event.key.keysym.sym == SDLK_LSHIFT ||
 				    event.key.keysym.sym == SDLK_RSHIFT)
 					shiftptr = 1;
@@ -155,7 +199,7 @@ int Main_Proc(void)
 				if (event.key.keysym.scancode - 8 >= 0)
 				{
 					//if (pressed[event.key.keysym.scancode - 8] != 2)
-						pressed[event.key.keysym.scancode - 8] = 1;
+					pressed[event.key.keysym.scancode - 8] = 1;
 					ProcessKeyBuf(event.key.keysym.sym);
 				}
 				break;
@@ -207,7 +251,6 @@ int Main_Proc(void)
 						MouseButton = MouseButton | event.button.button;
 						break;
 				}
-
 				break;
 
 			case SDL_MOUSEBUTTONUP:
@@ -215,132 +258,127 @@ int Main_Proc(void)
 					MouseButton & ~event.button.button;
 				break;
 
-		    case SDL_JOYHATMOTION: // POV hats act as direction pad
-			    if (event.jhat.hat == 0) // only support the first POV hat for now
-			    {
-				    switch (event.jhat.value)
-				    {
+			case SDL_JOYHATMOTION: // POV hats act as direction pad
+				offset = HatOffset[event.jhat.which];
+				if (offset >= (256 + 128 + 64)) break;
+				switch (event.jhat.value)
+			    	{
 				        case SDL_HAT_CENTERED:
-						    pressed[0x100 + event.jhat.which * 32 + 0] = 0;
-							pressed[0x100 + event.jhat.which * 32 + 1] = 0;
-							pressed[0x100 + event.jhat.which * 32 + 2] = 0;
-							pressed[0x100 + event.jhat.which * 32 + 3] = 0;
-							break;
-					    case SDL_HAT_UP:
-						    pressed[0x100 + event.jhat.which * 32 + 3] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 2] = 0;
-							break;
-					    case SDL_HAT_RIGHTUP:
-						    pressed[0x100 + event.jhat.which * 32 + 0] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 3] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 1] = 0;
-							pressed[0x100 + event.jhat.which * 32 + 2] = 0;
-							break;
-					    case SDL_HAT_RIGHT:
-						    pressed[0x100 + event.jhat.which * 32 + 0] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 1] = 0;
-							break;
-					    case SDL_HAT_RIGHTDOWN:
-						    pressed[0x100 + event.jhat.which * 32 + 0] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 2] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 1] = 0;
-							pressed[0x100 + event.jhat.which * 32 + 3] = 0;
-							break;
-					    case SDL_HAT_DOWN:
-						    pressed[0x100 + event.jhat.which * 32 + 2] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 3] = 0;
-							break;
-					    case SDL_HAT_LEFTDOWN:
-						    pressed[0x100 + event.jhat.which * 32 + 1] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 2] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 0] = 0;
-							pressed[0x100 + event.jhat.which * 32 + 3] = 0;
-							break;
-					    case SDL_HAT_LEFT:
-						    pressed[0x100 + event.jhat.which * 32 + 1] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 0] = 0;
-							break;
-					    case SDL_HAT_LEFTUP:
-						    pressed[0x100 + event.jhat.which * 32 + 1] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 3] = 1;
-							pressed[0x100 + event.jhat.which * 32 + 0] = 0;
-							pressed[0x100 + event.jhat.which * 32 + 2] = 0;
-							break;
-					}
+						pressed[offset] = 0;
+						pressed[offset + 1] = 0;
+						pressed[offset + 2] = 0;
+						pressed[offset + 3] = 0;
+						break;
+					case SDL_HAT_UP:
+						pressed[offset + 3] = 1;
+						pressed[offset + 2] = 0;
+						break;
+					case SDL_HAT_RIGHTUP:
+						pressed[offset] = 1;
+						pressed[offset + 3] = 1;
+						pressed[offset + 1] = 0;
+						pressed[offset + 2] = 0;
+						break;
+					case SDL_HAT_RIGHT:
+						pressed[offset] = 1;
+						pressed[offset + 1] = 0;
+						break;
+					case SDL_HAT_RIGHTDOWN:
+						pressed[offset] = 1;
+						pressed[offset + 2] = 1;
+						pressed[offset + 1] = 0;
+						pressed[offset + 3] = 0;
+						break;
+					case SDL_HAT_DOWN:
+						pressed[offset + 2] = 1;
+						pressed[offset + 3] = 0;
+						break;
+					case SDL_HAT_LEFTDOWN:
+						pressed[offset + 1] = 1;
+						pressed[offset + 2] = 1;
+						pressed[offset] = 0;
+						pressed[offset + 3] = 0;
+						break;
+					case SDL_HAT_LEFT:
+						pressed[offset + 1] = 1;
+						pressed[offset] = 0;
+						break;
+					case SDL_HAT_LEFTUP:
+						pressed[offset + 1] = 1;
+						pressed[offset + 3] = 1;
+						pressed[offset] = 0;
+						pressed[offset + 2] = 0;
+						break;
 				}
 				break;
 
 			/*
-			   joystick trackball code untested; change the test values
-			   if the motion is too sensitive (or not sensitive enough);
-			   only the first trackball is supported for now. we could get
-			   really general here, but this may break the format of 'pressed'
+			   joystick trackball code untested; change the test
+			   values if the motion is too sensitive (or not
+			   sensitive enough)
 			 */
 			case SDL_JOYBALLMOTION:
-			    //CurrentJoy = event.jball.which;
-				if (event.jball.ball == 0)
+				offset = BallOffset[event.jball.which];
+				offset += event.jball.ball;
+				if (offset >= (256 + 128 + 64)) break;
+				if (event.jball.xrel < -100)
 				{
-					if (event.jball.xrel < -100)
-					{
-						pressed[0x100 + event.jball.which * 32 + 6] = 0;
-						pressed[0x100 +	event.jball.which * 32 + 7] = 1;
-					}
-					if (event.jball.xrel > 100)
-					{
-						pressed[0x100 + event.jball.which * 32 + 6] = 1;
-						pressed[0x100 +	event.jball.which * 32 + 7] = 0;
-					}
-					if (event.jball.yrel < -100)
-					{
-						pressed[0x100 + event.jball.which * 32 + 8] = 0;
-						pressed[0x100 + event.jball.which * 32 + 9] = 1;
-					}
-					if (event.jball.yrel > 100)
-					{
-						pressed[0x100 + event.jball.which * 32 + 8] = 1;
-						pressed[0x100 + event.jball.which * 32 + 9] = 0;
+					pressed[offset] = 0;
+					pressed[offset + 1] = 1;
 				}
+				if (event.jball.xrel > 100)
+				{
+					pressed[offset] = 1;
+					pressed[offset + 1] = 0;
+				}
+				if (event.jball.yrel < -100)
+				{
+					pressed[offset + 2] = 0;
+					pressed[offset + 3] = 1;
+				}
+				if (event.jball.yrel > 100)
+				{
+					pressed[offset + 2] = 1;
+					pressed[offset + 3] = 0;
 				}
 				break;
 
 			case SDL_JOYAXISMOTION:
-				for (j = 0; j < 3; j++)
+				offset = AxisOffset[event.jaxis.which];
+				offset += event.jaxis.axis * 2;
+				if (offset >= (256 + 128 + 64)) break;
+//				printf("DEBUG axis offset: %d\n", offset);
+				if (event.jaxis.value < -16384)
 				{
-					if (event.jaxis.axis == j)
-					{
-						if (event.jaxis.value < -16384)
-						{
-							pressed[0x100 +
-								event.jaxis.which *	32 + 2 * j + 1] = 1;
-							pressed[0x100 +
-								event.jaxis.which * 32 + 2 * j + 0] = 0;
-						}
-						else if (event.jaxis.value > 16384)
-						{
-							pressed[0x100 +
-								event.jaxis.which * 32 + 2 * j + 0] = 1;
-							pressed[0x100 +
-								event.jaxis.which * 32 + 2 * j + 1] = 0;
-						}
-						else
-						{
-							pressed[0x100 +
-								event.jaxis.which * 32 + 2 * j + 0] = 0;
-							pressed[0x100 +
-								event.jaxis.which * 32 + 2 * j + 1] = 0;
-						}
-					}
+					pressed[offset + 1] = 1;
+					pressed[offset + 0] = 0;
+				}
+				else if (event.jaxis.value > 16384)
+				{
+					pressed[offset + 0] = 1;
+					pressed[offset + 1] = 0;
+				}
+				else
+				{
+					pressed[offset + 0] = 0;
+					pressed[offset + 1] = 0;
 				}
 				break;
 
 			case SDL_JOYBUTTONDOWN:
-				pressed[0x100 + event.jbutton.which * 32 + 16 +
-					event.jbutton.button] = 1;
+				offset = ButtonOffset[event.jbutton.which];
+				offset += event.jbutton.button;
+//				printf("DEBUG button offset: %d\n", offset);
+				if (offset >= (256 + 128 + 64)) break;
+				pressed[offset] = 1;
 				break;
 
 			case SDL_JOYBUTTONUP:
-				pressed[0x100 + event.jbutton.which * 32 + 16 +
-					event.jbutton.button] = 0;
+				offset = ButtonOffset[event.jbutton.which];
+				offset += event.jbutton.button;
+//				printf("DEBUG button offset: %d\n", offset);
+				if (offset >= (256 + 64)) break;
+				pressed[offset] = 0;
 				break;
 			case SDL_QUIT:
 				LinuxExit();
@@ -503,6 +541,11 @@ int InitSound(void)
 		return FALSE;
 	}
 
+	if (Buffer)
+		free(Buffer);
+	Buffer = NULL;
+	Buffer_len = 0;
+
 	PrevSoundQuality = SoundQuality;
 	PrevStereoSound = StereoSound;
 
@@ -520,11 +563,12 @@ int InitSound(void)
 	}
 
 	wanted.samples = samptab[SoundQuality] * 128 * wanted.channels;
+
 	wanted.format = AUDIO_S16LSB;
 	wanted.userdata = NULL;
 	wanted.callback = UpdateSound;
 
-	if (SDL_OpenAudio(&wanted, NULL) < 0)
+	if (SDL_OpenAudio(&wanted, &audiospec) < 0)
 	{
 		fprintf(stderr, "Sound init failed!\n");
 		fprintf(stderr, "freq: %d, channels: %d, samples: %d\n",
@@ -534,6 +578,10 @@ int InitSound(void)
 	}
 	SDL_PauseAudio(0);
 
+	Buffer_len = (audiospec.size * 2);
+	Buffer_len = (Buffer_len + 255) & ~255; /* Align to SPCSize */
+	Buffer = malloc(Buffer_len);
+
 	return TRUE;
 }
 
@@ -545,6 +593,8 @@ int ReInitSound(void)
 BOOL InitJoystickInput(void)
 {
 	int i, max_num_joysticks;
+	int num_axes, num_buttons, num_hats, num_balls;
+	int js_fail = 0;
 
 	for (i = 0; i < 5; i++)
 		JoystickInput[i] = NULL;
@@ -563,12 +613,51 @@ BOOL InitJoystickInput(void)
 	SDL_JoystickEventState(SDL_ENABLE);
 
 	if (max_num_joysticks > 5) max_num_joysticks = 5;
+
 	for (i = 0; i < max_num_joysticks; i++)
 	{
 		JoystickInput[i] = SDL_JoystickOpen(i);
-		printf("Joystick %i (%i Buttons): %s\n", i,
-		       SDL_JoystickNumButtons(JoystickInput[i]),
-		       SDL_JoystickName(i));
+		num_axes = SDL_JoystickNumAxes(JoystickInput[i]);
+		num_buttons = SDL_JoystickNumButtons(JoystickInput[i]);
+		num_hats = SDL_JoystickNumHats(JoystickInput[i]);
+		num_balls = SDL_JoystickNumBalls(JoystickInput[i]);
+		printf("Device %i %s\n", i, SDL_JoystickName(i));
+		printf("  %i axis, %i buttons, %i hats, %i balls\n",
+			num_axes, num_buttons, num_hats, num_balls);
+
+		if (js_fail)
+		{
+			printf("Warning: Joystick won't work.\n");
+			continue;
+		}
+		if (i == 0)
+		{
+			AxisOffset[0] = 256;		// After key data
+			ButtonOffset[0] = AxisOffset[0] + num_axes * 2;
+//			printf("ButtonOffset %d\n", ButtonOffset[0]);
+			HatOffset[0] = ButtonOffset[0] + num_buttons;
+//			printf("HatOffset %d\n", HatOffset[0]);
+			BallOffset[0] = HatOffset[0] + num_hats * 4;
+//			printf("BallOffset %d\n", BallOffset[0]);
+			if ((BallOffset[0] + num_balls * 4) >= (256 + 128 + 64))
+				js_fail = 1;
+		}
+		else
+		{
+			AxisOffset[i] = BallOffset[i - 1] +
+				SDL_JoystickNumBalls(JoystickInput[i - 1]);
+			ButtonOffset[i] = AxisOffset[i] + num_axes * 2;
+			HatOffset[i] = ButtonOffset[i] + num_buttons;
+			BallOffset[i] = HatOffset[i] + num_hats * 4;
+			if ((BallOffset[i] + num_balls * 4) >= (256 + 128 + 64))
+				js_fail = 1;
+
+		}
+		if (js_fail)
+		{
+			printf("Warning: Too many buttons, axes, hats and/or Balls!\n");
+			printf("Warning: Joystick won't work fully.\n");
+		}
 	}
 
 	return TRUE;
@@ -580,22 +669,6 @@ BOOL InitInput()
 	return TRUE;
 }
 
-int saybitdepth()
-{
-  int MyBitsPerPixel;
-  const SDL_VideoInfo *info;
-  SDL_Init(SDL_INIT_VIDEO);
-  info = SDL_GetVideoInfo();
-  MyBitsPerPixel = info->vfmt->BitsPerPixel;
-  switch (MyBitsPerPixel)
-    {
-    case 0: printf("Cannot detect bitdepth. On fbcon and svgalib this is normal.\nTrying to force 16 bpp.\n\n"); break;
-    case 16: break;
-    default: printf("You are running in %d bpp, but ZSNES is forcing 16 bpp.\nYou may experience poor performance and/or crashing.\n\n", MyBitsPerPixel); break;
-    }
-  return 0;
-}
-
 int startgame(void)
 {
 	int status;
@@ -611,10 +684,14 @@ int startgame(void)
 		sdl_state = vid_none;
 	}
 
+	sem_sleep_rdy();
+
 	if (sdl_state == vid_soft) sw_end();
 #ifdef __OPENGL__
 	else if (sdl_state == vid_gl) gl_end();
-	saybitdepth();
+
+	SDL_Init(SDL_INIT_VIDEO);
+
 	if (UseOpenGL)
 	{
 		status = gl_start(WindowWidth, WindowHeight, BitDepth, FullScreen);
@@ -635,6 +712,7 @@ void LinuxExit(void)
 	if (sdl_state != vid_null)
 	{
 		SDL_WM_GrabInput(SDL_GRAB_OFF);	// probably redundant
+		sem_sleep_die();
 		SDL_Quit();
 	}
 	exit(0);
@@ -658,8 +736,8 @@ void Start60HZ(void)
 		update_ticks_pc = UPDATE_TICKS_GAME;
 	}
 
-	start = SDL_GetTicks();
-	start2 = SDL_GetTicks();
+	start = sem_GetTicks();
+	start2 = sem_GetTicks();
 	T36HZEnabled = 0;
 	T60HZEnabled = 1;
 }
@@ -674,8 +752,8 @@ void Start36HZ(void)
 	update_ticks_pc2 = UPDATE_TICKS_UDP;
 	update_ticks_pc = UPDATE_TICKS_GUI;
 
-	start = SDL_GetTicks();
-	start2 = SDL_GetTicks();
+	start = sem_GetTicks();
+	start2 = sem_GetTicks();
 	T60HZEnabled = 0;
 	T36HZEnabled = 1;
 }
@@ -685,10 +763,39 @@ void Stop36HZ(void)
 	T36HZEnabled = 0;
 }
 
+void init_hqNx(void)
+{
+	DWORD color32;
+	DWORD *p;
+	int i, j, k, r, g, b, Y, u, v;
+
+	for(i = 0, p = BitConv32Ptr; i < 65536; i++, p++) {
+		color32=((i&0xF800)<<8)+
+		        ((i&0x07E0)<<5)+
+		        ((i&0x001F)<<3)+0xFF000000;
+
+		*p = color32;
+	}
+
+	for (i=0; i<32; i++)
+	for (j=0; j<64; j++)
+	for (k=0; k<32; k++) {
+		r = i << 3;
+		g = j << 2;
+		b = k << 3;
+		Y = (r + g + b) >> 2;
+		u = 128 + ((r - b) >> 2);
+		v = 128 + ((-r + 2*g -b)>>3);
+		RGBtoYUVPtr[(i << 11) + (j << 5) + k] = (Y<<16) + (u<<8) + v;
+	}
+}
+
 void initwinvideo(void)
 {
 	DWORD newmode = 0;
 
+	init_hqNx();
+
 	if (CurMode != cvidmode)
 	{
 		CurMode = cvidmode;
@@ -790,7 +897,7 @@ void CheckTimers(void)
 	int i;
 
 	//QueryPerformanceCounter((LARGE_INTEGER*)&end2);
-	end2 = SDL_GetTicks();
+	end2 = sem_GetTicks();
 
 	while ((end2 - start2) >= update_ticks_pc2)
 	{
@@ -810,9 +917,9 @@ void CheckTimers(void)
 	}
 
 	if (T60HZEnabled)
-{
+	{
 		//QueryPerformanceCounter((LARGE_INTEGER*)&end);
-		end = SDL_GetTicks();
+		end = sem_GetTicks();
 
 		while ((end - start) >= update_ticks_pc)
 		{
@@ -824,6 +931,7 @@ void CheckTimers(void)
 			   }
 			 */
 			Game60hzcall();
+			SDL_SemPost(sem_frames);
 			start += update_ticks_pc;
 		}
 	}
@@ -831,7 +939,7 @@ void CheckTimers(void)
 	if (T36HZEnabled)
 	{
 		//QueryPerformanceCounter((LARGE_INTEGER*)&end);
-		end = SDL_GetTicks();
+		end = sem_GetTicks();
 
 		while ((end - start) >= update_ticks_pc)
 		{
@@ -850,97 +958,111 @@ void CheckTimers(void)
 
 void UpdateSound(void *userdata, Uint8 * stream, int len)
 {
-	const int SPCSize = 256;
-	int DataNeeded;
-	int i;
-	Uint8 *ptr;
+	int left;
 
-	len /= 2;		/* only 16bit here */
+	left = Buffer_len - Buffer_head;
 
-	ptr = stream;
-	DataNeeded = len;
-
-	/* take care of the things we left behind last time */
-	if (BufferLeftOver)
-	{
-		DataNeeded -= BufferLeftOver;
-
-		memcpy(ptr, &Buffer[BufferLeftOver],
-		       (SPCSize - BufferLeftOver) * 2);
-
-		ptr += (SPCSize - BufferLeftOver) * 2;
-		BufferLeftOver = 0;
+	if (left <= len) {
+		memcpy(stream, &Buffer[Buffer_head], left);
+		stream += left;
+		len -= left;
+		Buffer_head = 0;
+		Buffer_fill -= left;
 	}
 
-	if (len & 255)
-	{			/* we'll save the rest first */
-		DataNeeded -= 256;
+	if (len) {
+		memcpy(stream, &Buffer[Buffer_head], len);
+		Buffer_head += len;
+		Buffer_fill -= len;
 	}
+}
 
-	while (DataNeeded > 0)
+void sem_sleep(void)
+{
+	end = update_ticks_pc - (sem_GetTicks() - start) - .2f;
+	if (end>0.f) SDL_SemWaitTimeout(sem_frames, (int)end);
+}
+
+static SDL_Thread *sem_threadid = NULL;
+static int sem_threadrun;
+
+int sem_thread(void *param)
+{
+	while (sem_threadrun)
 	{
-		SoundProcess();
-
-		for (i = 0; i < SPCSize; i++)
+		if (T60HZEnabled)
 		{
-			if (T36HZEnabled)
-			{
-				Buffer[i] = 0;
-			}
-			else
-			{
-				if (DSPBuffer[i] > 32767)
-					Buffer[i] = 32767;
-				else if (DSPBuffer[i] < -32767)
-					Buffer[i] = -32767;
-				else
-					Buffer[i] = DSPBuffer[i];
-			}
+			SDL_SemPost(sem_frames);
+			usleep(romispal ? 2000 : 1000);
 		}
-
-		memcpy(ptr, &Buffer[0], SPCSize * 2);
-		ptr += SPCSize * 2;
-
-		DataNeeded -= SPCSize;
+		else
+			usleep(20000);
 	}
+	return(0);
+}
 
-	if (DataNeeded)
+void sem_sleep_rdy(void)
+{
+	if (sem_frames) return;
+	sem_frames = SDL_CreateSemaphore(0);
+	sem_threadrun = 1;
+	sem_threadid = SDL_CreateThread(sem_thread, 0);
+}
+
+void sem_sleep_die(void)
+{
+	if (sem_threadid)
 	{
-		DataNeeded += 256;
-		BufferLeftOver = DataNeeded;
-
-		SoundProcess();
-
-		for (i = 0; i < SPCSize; i++)
-		{
-			if (T36HZEnabled)
-			{
-				Buffer[i] = 0;
-			}
-			else
-			{
-				if (DSPBuffer[i] > 32767)
-					Buffer[i] = 32767;
-				else if (DSPBuffer[i] < -32767)
-					Buffer[i] = -32767;
-				else
-					Buffer[i] = DSPBuffer[i];
-			}
-		}
-
-		memcpy(ptr, &Buffer[0], DataNeeded * 2);
+		sem_threadrun = 0;
+		SDL_WaitThread(sem_threadid, NULL);
+		sem_threadid = NULL;
+	}
+	if (sem_frames)
+	{
+		SDL_DestroySemaphore(sem_frames);
+		sem_frames = NULL;
 	}
 }
 
 void UpdateVFrame(void)
 {
-	/* rcg06172001 get menu animations running correctly... */
-	/*if (GUIOn2 == 1 && IsActivated == 0) SDL_WaitEvent(NULL);*/
-	if (GUIOn2 == 1 && IsActivated == 0)
-		SDL_Delay(100);  /* spare the CPU a little. */
+	const int SPCSize = 256;
+	int i;
 
 	CheckTimers();
 	Main_Proc();
+
+	/* Process sound */
+
+	/* take care of the things we left behind last time */
+	SDL_LockAudio();
+	while (Buffer_fill < Buffer_len) {
+		short *ptr = (short*)&Buffer[Buffer_tail];
+
+		SoundProcess();
+
+		for (i = 0; i < SPCSize; i++, ptr++)
+		{
+			if (T36HZEnabled)
+			{
+				*ptr = 0;
+			}
+			else
+			{
+				if (DSPBuffer[i] > 32767)
+					*ptr = 32767;
+				else if (DSPBuffer[i] < -32767)
+					*ptr = -32767;
+				else
+					*ptr = DSPBuffer[i];
+			}
+		}
+
+		Buffer_fill += SPCSize * 2;
+		Buffer_tail += SPCSize * 2;
+		if (Buffer_tail >= Buffer_len) Buffer_tail = 0;
+	}
+	SDL_UnlockAudio();
 }
 
 void clearwin()
@@ -1036,3 +1158,19 @@ void GetLocalTime()
 	SystemTimewMinute = timeptr->tm_min;
 	SystemTimewSecond = timeptr->tm_sec;
 }
+
+/* evul, maybe should use something other than constructor method */
+void __attribute__ ((stdcall, constructor)) sem_StartTicks()
+{
+	gettimeofday(&sem_start, NULL);
+}
+
+float sem_GetTicks()
+{
+	struct timeval now;
+	float ticks;
+
+	gettimeofday(&now, NULL);
+	ticks=((float)(now.tv_sec-sem_start.tv_sec))*1000.f+((float)(now.tv_usec-sem_start.tv_usec))*.001f;
+	return(ticks);
+}
