@@ -1,4 +1,4 @@
-//Copyright (C) 1997-2001 ZSNES Team ( zsknight@zsnes.com / _demo_@zsnes.com )
+//Copyright (C) 1997-2004 ZSNES Team ( zsknight@zsnes.com / _demo_@zsnes.com )
 //
 //This program is free software; you can redistribute it and/or
 //modify it under the terms of the GNU General Public License
@@ -30,10 +30,14 @@ extern void ConvertJoyMap1(void);
 extern void ConvertJoyMap2(void);
 extern void displayparams(void);
 extern void makeextension(void);
+extern unsigned char KitchenSync;
+extern unsigned char Force60hz;
+
+int pccmdline(void);
 
 extern unsigned char Palette0, SPC700sh, OffBy1Line, DSPDisable,
                      MMXSupport, Force8b, ForcePal, GUIClick, MouseDis,
-                     MusicRelVol, ScreenScale, SoundCompD, SoundQuality,
+                     MusicRelVol, ScreenScale, SoundQuality,
                      StereoSound, V8Mode, antienab, cvidmode, debugdisble,
                      debugger, enterpress, finterleave, frameskip,
                      gammalevel, guioff, per2exec, pl1contrl, pl2contrl,
@@ -42,6 +46,10 @@ extern unsigned char Palette0, SPC700sh, OffBy1Line, DSPDisable,
                      NetChatFirst,NetServer,NetNewNick,
                      NetFilename,CmdLineTCPIPAddress,NetQuitAfter,UDPConfig,
                      CmdLineNetPlay;
+                     
+// FIX STATMAT
+extern unsigned char autoloadstate;
+// FIX STATMAT                     
 
 void ccmdline(void);
 
@@ -65,13 +73,14 @@ char ucase(char ch){
 }
 
 extern HINSTANCE hInst;
-WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
+int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
 {
    int longueur;
 
    char path_buffer[_MAX_PATH], drive[_MAX_DRIVE] ,dir[_MAX_DIR];
    char fname2[_MAX_FNAME],ext[_MAX_EXT], File[_MAX_PATH];
-   DWORD dwRead;
+   //MK: Unused 2003/08/31
+   //DWORD dwRead;
 
    char * strp;
 
@@ -481,7 +490,7 @@ int pccmdline(void)
         {
           if(!hasroom) return 4;
           SoundQuality=my_atoi(argv[p+1]);
-          if(SoundQuality > 5) return 12;
+          if(SoundQuality > 6) return 12;
           p++;
           break;
         }
@@ -597,7 +606,15 @@ int pccmdline(void)
         }
         case 'z':
         {
-          StereoSound=1;
+          // FIX STATMAT
+          if(nn == 's')
+          {				
+            if(!hasroom) return 4;
+            autoloadstate=my_atoi(argv[p+1]) + 1;
+            p++;
+          }
+          else StereoSound=0;
+          // FIX STATMAT
           break;
         }
         case 'd':
@@ -614,11 +631,6 @@ int pccmdline(void)
 //          }
           break;
         }
-        case 'b':
-        {
-          SoundCompD=1;
-          break;
-        }
         case 'c':
         {
           if(nn == 'c')
@@ -661,6 +673,16 @@ int pccmdline(void)
           MouseDis=1;
           break;
         }
+        case 'K':
+        {
+          KitchenSync=1;
+          break;
+        }
+        case '6':
+        {
+          Force60hz=1;
+          break;
+        }
         case '?':
         {
           return 9;
